// **************************************************************************
// * This file is part of the zenXML project. It is distributed under the   *
// * Boost Software License: http://www.boost.org/LICENSE_1_0.txt           *
// * Copyright (C) Zenju (zenju AT gmx DOT de) - All Rights Reserved        *
// **************************************************************************

#ifndef ZEN_XML_PARSER_HEADER_81248670213764583021432
#define ZEN_XML_PARSER_HEADER_81248670213764583021432

#include <cstdio>
#include <zen/string_traits.h>
#include "dom.h"
#include "error.h"

namespace zen
{
/**
\file
\brief Convert an XML document object model (class XmlDoc) to and from a byte stream representation.
*/

///Save XML document as a byte stream
/**
\param doc Input XML document
\param lineBreak Line break, default: carriage return + new line
\param indent Indentation, default: four space characters
\return Output byte stream
*/
std::string serialize(const XmlDoc& doc,
                      const std::string& lineBreak = "\r\n",
                      const std::string& indent = "    "); //throw ()

///Exception thrown due to an XML parsing error
struct XmlParsingError : public XmlError
{
    XmlParsingError(size_t rowNo, size_t colNo) : row(rowNo), col(colNo) {}
    ///Input file row where the parsing error occured
    size_t row; //beginning with 1
    ///Input file column where the parsing error occured
    size_t col; //
};


///Load XML document from a byte stream
/**
\param stream Input byte stream
\param doc Output XML document
\throw XmlParsingError
*/
void parse(const std::string& stream, XmlDoc& doc); //throw XmlParsingError






























//---------------------------- implementation ----------------------------
namespace implementation
{
inline
std::pair<char, char> hexify(unsigned char c)
{
    auto hexifyDigit = [](int num) -> char //input [0, 15], output 0-9, A-F
    {
        assert(0 <= num && num <= 15); //guaranteed by design below!
        return static_cast<char>(num <= 9 ? //no signed/unsigned char problem here!
        '0' + num :
        'A' + (num - 10));
    };
    return std::make_pair(hexifyDigit(c / 16), hexifyDigit(c % 16));
}


inline
char unhexify(char high, char low)
{
    auto unhexifyDigit = [](char hex) -> int //input 0-9, a-f, A-F; output range: [0, 15]
    {
        if ('0' <= hex && hex <= '9') //no signed/unsigned char problem here!
            return hex - '0';
        else if ('A' <= hex && hex <= 'F')
            return (hex - 'A') + 10;
        else if ('a' <= hex && hex <= 'f')
            return (hex - 'a') + 10;
        assert(false);
        return 0;
    };
    return static_cast<unsigned char>(16 * unhexifyDigit(high) + unhexifyDigit(low)); //[!] convert to unsigned char first, then to char (which may be signed)
};


template <class Predicate> inline
std::string normalize(const std::string& str, Predicate pred) //pred: unary function taking a char, return true if value shall be encoded as hex
{
    std::string output;
    std::for_each(str.begin(), str.end(),
                  [&](char c)
    {
        if (c == '&')
            output += "&amp;";
        else if (c == '\'')
            output += "&apos;";
        else if (c == '<')
            output += "&lt;";
        else if (c == '>')
            output += "&gt;";
        else if (c == '\"')
            output += "&quot;";
        else if (pred(c))
        {
            output += "&#x";
            const auto hexDigits = hexify(c);
            output += hexDigits.first;
            output += hexDigits.second;
            output += ';';
            //hexify beats "printNumber<std::string>("&#x%02X;", c)" by a nice factor of 3!
        }
        else
            output += c;
    });

    return output;
}

inline
std::string normalizeName(const std::string& str)
{
    return normalize(str, [](char ch) { return isWhiteSpace(ch) || ch == '=' || ch == '/'; });
}

inline
std::string normalizeValue(const std::string& str)
{
    return normalize(str, [](char ch) { return static_cast<unsigned char>(ch) < 32; });
}


namespace
{
std::string denormalize(const std::string& str)
{
    std::string output;
    for (auto iter = str.begin(); iter != str.end(); ++iter)
    {
        const char c = *iter;

        if (c == '&')
        {
            auto checkEntity = [&](const char* placeholder, char realVal) -> bool
            {
                size_t strLen = strLength(placeholder);

                if (str.end() - iter >= static_cast<int>(strLen) && std::equal(iter, iter + strLen, placeholder))
                {
                    output += realVal;
                    iter += strLen - 1;
                    return true;
                }
                return false;
            };

            if (checkEntity("&amp;", '&'))
                continue;
            if (checkEntity("&apos;", '\''))
                continue;
            if (checkEntity("&lt;", '<'))
                continue;
            if (checkEntity("&gt;", '>'))
                continue;
            if (checkEntity("&quot;", '\"'))
                continue;

            if (str.end() - iter >= 6 &&
                iter[1] == '#' &&
                iter[2] == 'x' &&
                iter[5] == ';')
            {
                output += unhexify(iter[3], iter[4]);
                iter += 5;
                continue;
                //unhexify beats "::sscanf(&iter[3], "%02X", &tmp)" by a factor of 3000 for ~250000 calls!!!
            }

            output += c; //unexpected char!
        }
        else
            output += c;
    };

    return output;
}


void serialize(const XmlElement& element, std::string& stream,
               const std::string& lineBreak,
               const std::string& indent,
               size_t indentLevel)
{
    const std::string& nameFmt = normalizeName(element.getNameAs<std::string>());

    for (size_t i = 0; i < indentLevel; ++i)
        stream += indent;

    stream += '<' + nameFmt;

    auto attr = element.getAttributes();
    for (auto iter = attr.first; iter != attr.second; ++iter)
        stream += ' ' + normalizeName(iter->first) + "=\"" + normalizeValue(iter->second) + "\"";

    //no support for mixed-mode content
    auto iterPair = element.getChildren();
    if (iterPair.first != iterPair.second) //structured element
    {
        stream += '>' + lineBreak;

        std::for_each(iterPair.first, iterPair.second,
        [&](const XmlElement & el) { serialize(el, stream, lineBreak, indent, indentLevel + 1); });

        for (size_t i = 0; i < indentLevel; ++i)
            stream += indent;
        stream += "</" + nameFmt + '>' + lineBreak;
    }
    else
    {
        std::string value;
        element.getValue(value);

        if (!value.empty()) //value element
            stream += '>' + normalizeValue(value) + "</" + nameFmt + '>' + lineBreak;
        else //empty element
            stream += "/>" + lineBreak;
    }
}

std::string serialize(const XmlDoc& doc,
                      const std::string& lineBreak,
                      const std::string& indent)
{
    std::string version = doc.getVersionAs<std::string>();
    if (!version.empty())
        version = " version=\"" + normalizeValue(version) + "\"";

    std::string encoding = doc.getEncodingAs<std::string>();
    if (!encoding.empty())
        encoding = " encoding=\"" + normalizeValue(encoding) + "\"";

    std::string standalone = doc.getStandaloneAs<std::string>();
    if (!standalone.empty())
        standalone = " standalone=\"" + normalizeValue(standalone) + "\"";

    std::string output = "<?xml" + version + encoding + standalone + "?>" + lineBreak;
    serialize(doc.root(), output, lineBreak, indent, 0);
    return output;
}
}
}

inline
std::string serialize(const XmlDoc& doc,
                      const std::string& lineBreak,
                      const std::string& indent) { return implementation::serialize(doc, lineBreak, indent); }

/*
Grammar for XML parser
-------------------------------
document-expression:
	<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
    element-expression:

element-expression:
	<string attributes-expression/>
	<string attributes-expression> pm-expression </string>

element-list-expression:
	<empty>
	element-expression element-list-expression

attributes-expression:
	<empty>
	string="string" attributes-expression

pm-expression:
	string
    element-list-expression
*/

class XmlParser
{
public:
    XmlParser(const std::string& stream) :
        scn(stream),
        tk(scn.nextToken()) {}

    void parse(XmlDoc& doc) //throw XmlParsingError
    {
        //declaration (optional)
        if (token().type == Token::TK_DECL_BEGIN)
        {
            nextToken();

            while (token().type == Token::TK_TEXT)
            {
                std::string attribName = token().text;
                nextToken();

                consumeToken(Token::TK_EQUAL);

                if (token().type != Token::TK_QUOTE)
                    throw XmlParsingError(scn.position().first, scn.position().second);
                std::string attribValue = scn.extractRawText();
                nextToken();

                consumeToken(Token::TK_QUOTE);

                if (attribName == "version")
                    doc.setVersion(attribValue);
                else if (attribName == "encoding")
                    doc.setEncoding(attribValue);
                else if (attribName == "standalone")
                    doc.setStandalone(attribValue);
            }

            consumeToken(Token::TK_DECL_END);
        }

        XmlDoc dummy;
        parseChildElements(dummy.root());

        auto iterPair = dummy.root().getChildren();
        if (iterPair.first != iterPair.second)
            doc.root().swap(*iterPair.first);

        consumeToken(Token::TK_END);
    };

private:
    XmlParser(const XmlParser&);
    XmlParser& operator=(const XmlParser&);

    struct Token
    {
        enum Type
        {
            TK_LESS,
            TK_GREATER,
            TK_LESS_SLASH,
            TK_SLASH_GREATER,
            TK_EQUAL,
            TK_QUOTE,
            TK_DECL_BEGIN,
            TK_DECL_END,
            TK_TEXT,
            TK_END
        };

        Token(Type t) : type(t) {}

        Type type;
        std::string text; //if type == TK_TEXT
    };

    class Scanner
    {
    public:
        Scanner(const std::string& phrase) : stream(phrase), pos(stream.begin())
        {
            if (zen::startsWith(phrase, BYTE_ORDER_MARK_UTF8))
                pos += 3;

            tokens.push_back(std::make_pair("<?xml", Token::TK_DECL_BEGIN));
            tokens.push_back(std::make_pair("?>",    Token::TK_DECL_END));
            tokens.push_back(std::make_pair("</", Token::TK_LESS_SLASH));
            tokens.push_back(std::make_pair("/>", Token::TK_SLASH_GREATER));
            tokens.push_back(std::make_pair("<" , Token::TK_LESS));
            tokens.push_back(std::make_pair(">" , Token::TK_GREATER));
            tokens.push_back(std::make_pair("=" , Token::TK_EQUAL));
            tokens.push_back(std::make_pair("\"", Token::TK_QUOTE));
            tokens.push_back(std::make_pair("\'", Token::TK_QUOTE));
        }

        Token nextToken()
        {
            //skip whitespace
            pos = std::find_if(pos, stream.end(), [](char c) { return !zen::isWhiteSpace(c); });

            if (pos == stream.end())
                return Token(Token::TK_END);

            for (TokenList::const_iterator i = tokens.begin(); i != tokens.end(); ++i)
                if (startsWith(pos, i->first))
                {
                    pos += i->first.size();
                    return Token(i->second);
                }

            auto textEnd = std::find_if(pos, stream.end(), [](char c)
            {
                return c == '<'  ||
                       c == '>'  ||
                       c == '\'' ||
                       c == '\"' ||
                       c == '='  ||
                       c == '/'  ||
                       zen::isWhiteSpace(c);
            });

            auto letterCount = textEnd - pos;
            if (letterCount != 0)
            {
                Token out(Token::TK_TEXT);
                out.text = implementation::denormalize(std::string(&*pos, letterCount));
                pos = textEnd;
                return out;
            }

            throw XmlParsingError(position().first, position().second); //cannot arrive here
        }

        std::string extractRawText()
        {
            auto iter = std::find_if(pos, stream.end(), [](char c)
            {
                return c == '<'  ||
                       c == '>'  ||
                       c == '\'' ||
                       c == '\"';
            });
            std::string output(&*pos, iter - pos);
            pos = iter;
            return implementation::denormalize(output);
        }

        std::pair<size_t, size_t> position() const //current (row/col) beginning with 1
        {
            //seek last line break
            std::string::const_iterator iter = pos;
            while (iter != stream.begin() && *iter != '\n')
                --iter;

            return std::make_pair(std::count(stream.begin(), pos, '\n') + 1, pos - iter);
        }

    private:
        Scanner(const Scanner&);
        Scanner& operator=(const Scanner&);

        bool startsWith(std::string::const_iterator iter, const std::string& prefix) const
        {
            if (stream.end() - iter < static_cast<int>(prefix.size()))
                return false;
            return std::equal(prefix.begin(), prefix.end(), iter);
        }

        typedef std::vector<std::pair<std::string, Token::Type> > TokenList;
        TokenList tokens;

        const std::string stream;
        std::string::const_iterator pos;
    };

    void parseChildElements(XmlElement& parent)
    {
        while (token().type == Token::TK_LESS)
        {
            consumeToken(Token::TK_LESS);

            if (token().type != Token::TK_TEXT)
                throw XmlParsingError(scn.position().first, scn.position().second);
            std::string elementName = token().text;
            nextToken();

            XmlElement& newElement = parent.addChild(elementName);

            parseAttributes(newElement);

            if (token().type == Token::TK_SLASH_GREATER) //empty element
            {
                nextToken();
                continue;
            }

            if (token().type != Token::TK_GREATER)
                throw XmlParsingError(scn.position().first, scn.position().second);
            std::string elementValue = scn.extractRawText();
            nextToken();

            //no support for mixed-mode content
            if (token().type == Token::TK_LESS) //structured element
                parseChildElements(newElement);
            else //value element
                newElement.setValue(elementValue);


            consumeToken(Token::TK_LESS_SLASH);

            if (token().type != Token::TK_TEXT ||
                elementName != token().text)
                throw XmlParsingError(scn.position().first, scn.position().second);
            nextToken();

            consumeToken(Token::TK_GREATER);
        }
    };

    void parseAttributes(XmlElement& element)
    {
        while (token().type == Token::TK_TEXT)
        {
            std::string attribName = token().text;
            nextToken();

            consumeToken(Token::TK_EQUAL);

            if (token().type != Token::TK_QUOTE)
                throw XmlParsingError(scn.position().first, scn.position().second);
            std::string attribValue = scn.extractRawText();
            nextToken();

            consumeToken(Token::TK_QUOTE);

            element.setAttribute(attribName, attribValue);
        }
    }


    const Token& token() const { return tk; }
    void nextToken() { tk = scn.nextToken(); }
    void consumeToken(Token::Type t)
    {
        if (token().type != t)
            throw XmlParsingError(scn.position().first, scn.position().second);
        nextToken();
    }

    Scanner scn;
    Token tk;
};


inline
void parse(const std::string& stream, XmlDoc& doc) //throw XmlParsingError
{
    XmlParser parser(stream);
    parser.parse(doc);  //throw XmlParsingError
}
}

#endif //ZEN_XML_PARSER_HEADER_81248670213764583021432
